/// <reference path="./includes/jquery.d.ts" />
/// <reference path="./includes/require.d.ts" />

/* 
	Board Game Engine 
*/

/* Possible Difficulty Settings */
export enum Difficulty {
	Easy,
	Medium,
	Hard
}

export module Events { 
	export class PlayerMove {
		public player: Player;
		public from: BoardPlace;
		public to: BoardPlace;
		constructor(player: Player, from: BoardPlace, to: BoardPlace) {
			this.player = player;
			this.from = from;
			this.to = to; 
		}
	}
}


/* Player Class */
export class Player {
	public Name: string;
	public Board: GameBoard;
	public Ready : Boolean = false;
	private IsComputer : boolean = false;
	private Difficulty : Difficulty = Difficulty.Easy;

	constructor(name: string, board: GameBoard) {
		this.Name = name;
		this.Board = board;
	}

	public Place(piece: GamePiece, place: BoardPlace) {
		piece.Player = this;
		piece.Place(this.Board, place);
		this.fire("player:placed", new Events.PlayerMove( 
			this,
			null,
			place
		));
	}
	public Move(piece: GamePiece, place: BoardPlace) {
		if (piece.Player != this) throw "This player does not own this piece.";
		piece.Move(place);
		this.fire("player:moved", new Events.PlayerMove( 
			this,
			null,
			place
		));
	}
	public name(name?: string) : any {
		if(name===undefined) return this.Name;
		this.Name = name;
		this.fire("player:namechange", this);
		return this;
	}
	public ready(ready? :boolean) : any {
		if(ready===undefined) return this.Ready;
		this.Ready = ready;
		this.fire("player:ready", this);
		return this;
	}
	public TakeTurn(place : BoardPlace) { }
	public AutoMove() { }

	public on(eventname: string, action: any) { this.Board.on(eventname, action); }
	public fire(eventname: string, data:any=undefined) { this.Board.fire(eventname, data); }
}

export class GamePiece {
	public Type: string;
	public Board: GameBoard;
	public Player: Player;
	public $: JQuery;

	constructor(type: string) {
		this.Type = type;
		this.$ = $(document.createElement("div"));
	}
	public Place(board: GameBoard, place:BoardPlace) {
		place.Hold(this);
	}
	public Move(place: BoardPlace) {
		if ( this.Board === undefined) throw "You cannot move a peice not on a board.";
		place.Hold(this);
	}

	public appendTo($:JQuery) : GamePiece { this.$.appendTo($); return this; }
	public prependTo($:JQuery) : GamePiece { this.$.prependTo($); return this; }
	public css(style: {[stylename:string]:{any}}) : GamePiece { this.$.css(style); return this; }
	public style(classname: string) : GamePiece { this.$.addClass(classname); return this; }		/* Adds a css class for styling */
	public remove() : GamePiece { this.$.remove(); this.Board = undefined; return this; }
	public click(func? : Function) : GamePiece { this.$.click(func); return this; }
	public on(eventname: string, action: any) { this.Board.on(eventname, action); }
	public fire(eventname: string, data:any=undefined) { this.Board.fire(eventname, data); }
}


export class BoardPlace {
	public Board: GameBoard;
	public Holds : GamePiece;
	public Position: number;
	public $: JQuery;

	constructor(board: GameBoard) {
		this.Board = board;
		this.$ = $(document.createElement("div"))
			.appendTo(this.Board.$);
	}

	public Holding() : boolean {
		if (this.Holds === undefined) return false; return true;
	}
	public Hold(piece:GamePiece) : BoardPlace {
		if(this.Holding()) throw "This game place is already holding a piece.";
		this.Holds = piece;
		this.Holds.appendTo(this.$);
		return this;
	}
	public Release() {
		if (this.Holding()) this.Holds.remove();
		this.Holds = undefined;
	}

	public css(style: {[stylename:string]:{any}}) : BoardPlace { this.$.css(style); return this; }
	public style(classname: string) : BoardPlace { this.$.addClass(classname); return this; }		/* Adds a css class for styling */
	public remove() : BoardPlace { this.$.remove(); return this; }
	public click(func? : Function) : BoardPlace { this.$.click(func); return this; }
	public on(eventname: string, action: any) { this.Board.on(eventname, action); }
	public fire(eventname: string, data:any=undefined) { this.Board.fire(eventname, data); }
}

export class GameBoard {
	public Game : Game;
	public Places: BoardPlace[] = [];
	public $: JQuery;

	constructor(game : Game) {
		this.$ = $(document.createElement("div"));
		this.Game = game;
		this.AddPlaces();
	}

	public AddPlaces() { }		/* For overriding */
	public AddPlace(place:BoardPlace) : GameBoard { this.Places.push(place); place.Position = this.Places.length; return this; }
	public GetPlaceIndex(givenplace:BoardPlace) : number { 
		return givenplace.Position;
	}

	public appendTo($:JQuery) : GameBoard { this.$.appendTo($); return this; }
	public prependTo($:JQuery): GameBoard  { this.$.prependTo($); return this; }
	public css(style: {[stylename:string]:{any}}) : GameBoard { this.$.css(style); return this; }
	public style(classname: string) : GameBoard { this.$.addClass(classname); return this; }		/* Adds a css class for styling */
	public remove() : GameBoard { this.$.remove(); return this; }
	public click(func? : Function) : GameBoard { this.$.click(func); return this; }
	public on(eventname: string, action: any) { this.Game.on(eventname, action); }
	public fire(eventname: string, data:any=undefined) { this.Game.fire(eventname, data); }

}

export class Game {
	public Board: GameBoard;
	public Players : Player[] = [];	// A list of game players
	public _Player : Player;	 	// The current player whos turn it is
	public Started : boolean = false;
	public $ : JQuery;

	constructor(board: GameBoard, element : JQuery) {
		this.$ = $(document.createElement("div")).appendTo(element);
		this.Board = board.appendTo(this.$);
	}
	public AddPlayer(player: Player) : Game { this.Players.push(player); this._Player = player; this.fire("player:added", player); return this; }
	public RemoveAllPlayers() { this.Players = []; this.fire("player:clear"); }
	public CurrentPlayer(player?: Player) : any {
		if (player===undefined) {
			if (this._Player===undefined) throw "No players have been added to this game."; 
			return this._Player;
		}
		this._Player = player;
	}
	public Start() {
		$.each(this.Players, (index: number, player: Player)=> {
			if (!player.ready()) throw "Not all players are ready.";
		});
		this.Started = true;
		this.fire("game:started");
	}

	public on(eventname: string, action: any) { this.$.on(eventname, action); }
	public fire(eventname: string, data:any=undefined) { this.$.trigger(eventname, data); }

	/* Makes the current player take their turn */
	public TakeTurn(place: BoardPlace) {
		if (this.Started==false) this.Start();
		this.CurrentPlayer().TakeTurn(place);

		var player : Player = this.CheckWin();
		if (!player === undefined) this.Win(player);
	}

	/* 
		Moves the player onto the next one (for example after taking a turn 
		Default behaviour is to move on to the next player in the list - or the first if at the end of the list.
	*/
	public CyclePlayers() {
		var playernumber : number = 0;
		for (var i : number = 0; i < this.Players.length; i++) {
			if (this.Players[i] === this._Player) {
				playernumber = i;
				break;
			}
		}
		playernumber = playernumber+1;
		if (!this.Players[playernumber]) this._Player = this.Players[0];
		else this._Player = this.Players[playernumber];	
		
		this.fire("player:yourturn", this._Player);
	}

	public CheckWin() : Player { return undefined; }
	public Win(player : Player) { console.log(["win", player]); this.fire("game:win", player); }
	public Draw() { this.fire("game:draw"); }
}